"""titiler-image dependencies."""

import json
from dataclasses import dataclass
from typing import List, Optional

import httpx
from cachetools import TTLCache, cached
from fastapi import HTTPException, Query
from geojson_pydantic import MultiPolygon, Polygon
from rasterio.control import GroundControlPoint
from rio_tiler.types import RIOResampling
from typing_extensions import Annotated

from titiler.core.dependencies import DefaultDependency


@dataclass
class DatasetParams(DefaultDependency):
    """Dataset Optional parameters."""

    unscale: Annotated[
        Optional[bool],
        Query(
            title="Apply internal Scale/Offset",
            description="Apply internal Scale/Offset",
        ),
    ] = False
    resampling_method: Annotated[
        RIOResampling,
        Query(
            alias="resampling",
            description="Resampling method.",
        ),
    ] = "nearest"


@cached(TTLCache(maxsize=512, ttl=3600))
def get_gcps(gcps_file: str) -> List[GroundControlPoint]:
    """Fetch and parse GCPS file."""
    if gcps_file.startswith("http"):
        body = httpx.get(gcps_file).json()

    else:
        with open(gcps_file, "r") as f:
            body = json.load(f)

    return [
        # GroundControlPoint(row, col, x, y, z)
        # https://github.com/allmaps/iiif-api/blob/georef/source/extension/georef/index.md#35-the-resourcecoords-property
        GroundControlPoint(
            f["properties"]["resourceCoords"][1],
            f["properties"]["resourceCoords"][0],
            *f["geometry"]["coordinates"],  # x, y, z
            id=f.get("id"),
        )
        for f in body["features"]
    ]


@cached(TTLCache(maxsize=512, ttl=3600))
def get_cutline(cutline_file: str) -> str:
    """Fetch and parse Cutline file."""
    if cutline_file.startswith("http"):
        body = httpx.get(cutline_file).json()

    else:
        with open(cutline_file, "r") as f:
            body = json.load(f)

    # We assume the geojson is a Feature (not a Feature Collectionw)
    if "geometry" in body:
        body = body["geometry"]

    geom_type = body["type"]
    if geom_type == "Polygon":
        return Polygon.parse_obj(body).wkt

    elif geom_type == "MultiPolygon":
        return MultiPolygon.parse_obj(body).wkt

    else:
        raise HTTPException(
            status_code=400, detail=f"Invalid GeoJSON type: {geom_type}."
        )


@dataclass
class GCPSParams(DefaultDependency):
    """GCPS parameters."""

    gcps: Optional[List[GroundControlPoint]] = None
    cutline: Optional[str] = None
    gcps_order: Optional[int] = None

    def __init__(
        self,
        gcps: Annotated[
            Optional[List[str]],
            Query(
                title="Ground Control Points",
                description="Ground Control Points in form of `row (y), col (x), lon, lat, alt`",
            ),
        ] = None,
        gcps_file: Annotated[
            Optional[str],
            Query(title="Ground Control Points GeoJSON path"),
        ] = None,
        gcps_order: Annotated[
            Optional[int],
            Query(
                title="The maximum order to use for GCP derived polynomials if possible. The default is to autoselect based on the number of GCPs. A value of -1 triggers use of Thin Plate Spline instead of polynomials."
            ),
        ] = None,
        cutline: Annotated[
            Optional[str],
            Query(
                title="WKT Image Cutline (equivalent of the SVG Selector)",
                description="WKT Polygon or MultiPolygon.",
            ),
        ] = None,
        cutline_file: Annotated[
            Optional[str],
            Query(title="GeoJSON file for cutline"),
        ] = None,
    ):
        """Initialize GCPSParams and Cutline

        Note: We only want `gcps` or `cutline` to be forwarded to the reader so we use a custom `__init__` method used by FastAPI to parse the QueryParams.
        """
        if gcps:
            self.gcps: List[GroundControlPoint] = [  # type: ignore
                # WARNING: gpcs should be in form of `row (y), col (x), lon, lat, alt`
                GroundControlPoint(*list(map(float, gcps.split(","))))
                for gcps in gcps
            ]
        elif gcps_file:
            self.gcps = get_gcps(gcps_file)

        if gcps_order is not None:
            self.gcps_order = gcps_order

        if self.gcps and len(self.gcps) < 3:
            raise HTTPException(
                status_code=400, detail="Need at least 3 gcps to wrap an image."
            )

        if cutline:
            self.cutline = cutline

        elif cutline_file:
            self.cutline = get_cutline(cutline_file)



"""API settings."""

from typing import Optional

from pydantic import field_validator, model_validator
from pydantic_settings import BaseSettings


class ApiSettings(BaseSettings):
    """API settings"""

    name: str = "titiler-image"
    cors_origins: str = "*"
    cachecontrol: str = "public, max-age=3600"
    root_path: str = ""

    model_config = {
        "env_prefix": "TITILER_IMAGE_API_",
        "env_file": ".env",
        "extra": "ignore",
    }

    @field_validator("cors_origins")
    def parse_cors_origin(cls, v):
        """Parse CORS origins."""
        return [origin.strip() for origin in v.split(",")]


class IIIFSettings(BaseSettings):
    """IIIF settings"""

    # The maximum width in pixels supported for this image.
    max_width: Optional[int] = None

    # The maximum height in pixels supported for this image.
    max_height: Optional[int] = None

    # The maximum area in pixels supported for this image.
    max_area: Optional[int] = None

    model_config = {
        "env_prefix": "TITILER_IMAGE_IIIF_",
        "env_file": ".env",
        "extra": "ignore",
    }

    @model_validator(mode="before")
    def check_max(cls, values):
        """Check MaxWitdh and MaxHeight configuration."""
        # maxWidth must be specified if maxHeight is specified.
        keys = {"max_width", "max_height"}
        if keys.intersection(values):
            if "max_width" not in values:
                raise Exception("max_width has to be set if max_height is.")
            if "max_height" not in values:
                values["max_height"] = values["max_width"]

        return values


iiif_settings = IIIFSettings()
api_settings = ApiSettings()



"""Reader with GCPS support."""

import warnings
import xml.etree.ElementTree as ET
from typing import List, Optional, Union

import attr
import rasterio
from rasterio._path import _parse_path
from rasterio.control import GroundControlPoint
from rasterio.crs import CRS
from rasterio.dtypes import _gdal_typename
from rasterio.enums import MaskFlags
from rasterio.io import DatasetReader
from rasterio.transform import from_gcps
from rasterio.vrt import WarpedVRT
from rio_tiler import io
from rio_tiler.constants import WGS84_CRS
from rio_tiler.errors import NoOverviewWarning
from rio_tiler.utils import has_alpha_band


@attr.s
class Reader(io.Reader):
    """GCPS + Image Reader"""

    gcps: Optional[List[GroundControlPoint]] = attr.ib(default=None)
    gcps_crs: Optional[CRS] = attr.ib(default=WGS84_CRS)
    gcps_order: Optional[int] = attr.ib(default=None)

    cutline: Optional[str] = attr.ib(default=None)

    dataset: Union[DatasetReader, WarpedVRT] = attr.ib(init=False)

    def __attrs_post_init__(self):
        """Define _kwargs, open dataset and get info."""
        dataset = self._ctx_stack.enter_context(rasterio.open(self.input))

        # when external GCPS we create a VRT
        if self.gcps:
            vrt_xml = vrt_doc(dataset, gcps=self.gcps, gcps_crs=self.gcps_crs)
            dataset = self._ctx_stack.enter_context(rasterio.open(vrt_xml))

        vrt_options = {}

        # Options 1: GCPS (internal or external)
        if dataset.gcps[0]:
            vrt_options.update(
                {
                    "src_crs": dataset.gcps[1],
                    "src_transform": from_gcps(dataset.gcps[0]),
                }
            )
            if self.gcps_order is not None:
                vrt_options.update({"max_gcp_order": self.gcps_order})

        # Option 2: Cutline
        if self.cutline:
            vrt_options.update({"cutline": self.cutline})

        if vrt_options:
            vrt_options.update({"add_alpha": True})
            if dataset.nodata is not None:
                vrt_options.update(
                    {
                        "nodata": dataset.nodata,
                        "add_alpha": False,
                        "src_nodata": dataset.nodata,
                    }
                )

            elif has_alpha_band(dataset):
                vrt_options.update({"add_alpha": False})

            self.dataset = self._ctx_stack.enter_context(
                WarpedVRT(dataset, **vrt_options)
            )

        else:
            self.dataset = dataset

        self.bounds = tuple(self.dataset.bounds)
        self.crs = self.dataset.crs

        if self.colormap is None:
            self._get_colormap()

        if min(
            self.dataset.width, self.dataset.height
        ) > 512 and not self.dataset.overviews(1):
            warnings.warn(
                "The dataset has no Overviews. rio-tiler performances might be impacted.",
                NoOverviewWarning,
            )


def vrt_doc(  # noqa: C901
    src_dataset,
    gcps: Optional[List[GroundControlPoint]] = None,
    gcps_crs: Optional[CRS] = WGS84_CRS,
):
    """Make a VRT XML document.

    Adapted from rasterio.vrt._boundless_vrt_doc function
    """
    vrtdataset = ET.Element("VRTDataset")
    vrtdataset.attrib["rasterYSize"] = str(src_dataset.height)
    vrtdataset.attrib["rasterXSize"] = str(src_dataset.width)

    tags = src_dataset.tags()
    if tags:
        metadata = ET.SubElement(vrtdataset, "Metadata")
        for key, value in tags.items():
            v = ET.SubElement(metadata, "MDI")
            v.attrib["Key"] = key
            v.text = str(value)

    im_tags = src_dataset.tags(ns="IMAGE_STRUCTURE")
    if im_tags:
        metadata = ET.SubElement(vrtdataset, "Metadata")
        for key, value in im_tags.items():
            if key == "LAYOUT" and value == "COG":
                continue
            v = ET.SubElement(metadata, "MDI")
            v.attrib["Key"] = key
            v.text = str(value)

    if src_dataset.crs:
        srs = ET.SubElement(vrtdataset, "SRS")
        srs.text = src_dataset.crs.wkt if src_dataset.crs else ""
        geotransform = ET.SubElement(vrtdataset, "GeoTransform")
        geotransform.text = ",".join([str(v) for v in src_dataset.transform.to_gdal()])

    nodata_value = src_dataset.nodata

    if gcps:
        gcp_list = ET.SubElement(vrtdataset, "GCPList")
        gcp_list.attrib["Projection"] = str(gcps_crs)
        for gcp in gcps:
            g = ET.SubElement(gcp_list, "GCP")
            g.attrib["Id"] = gcp.id
            g.attrib["Pixel"] = str(gcp.col)
            g.attrib["Line"] = str(gcp.row)
            g.attrib["X"] = str(gcp.x)
            g.attrib["Y"] = str(gcp.y)

    for bidx, ci, block_shape, dtype in zip(
        src_dataset.indexes,
        src_dataset.colorinterp,
        src_dataset.block_shapes,
        src_dataset.dtypes,
    ):
        vrtrasterband = ET.SubElement(vrtdataset, "VRTRasterBand")
        vrtrasterband.attrib["dataType"] = _gdal_typename(dtype)
        vrtrasterband.attrib["band"] = str(bidx)

        if nodata_value is not None:
            nodata = ET.SubElement(vrtrasterband, "NoDataValue")
            nodata.text = str(nodata_value)

        colorinterp = ET.SubElement(vrtrasterband, "ColorInterp")
        colorinterp.text = ci.name.capitalize()

        source = ET.SubElement(vrtrasterband, "SimpleSource")
        sourcefilename = ET.SubElement(source, "SourceFilename")
        sourcefilename.attrib["relativeToVRT"] = "0"
        sourcefilename.text = _parse_path(src_dataset.name).as_vsi()
        sourceband = ET.SubElement(source, "SourceBand")
        sourceband.text = str(bidx)
        sourceproperties = ET.SubElement(source, "SourceProperties")
        sourceproperties.attrib["RasterXSize"] = str(src_dataset.width)
        sourceproperties.attrib["RasterYSize"] = str(src_dataset.height)
        sourceproperties.attrib["dataType"] = _gdal_typename(dtype)
        sourceproperties.attrib["BlockYSize"] = str(block_shape[0])
        sourceproperties.attrib["BlockXSize"] = str(block_shape[1])
        srcrect = ET.SubElement(source, "SrcRect")
        srcrect.attrib["xOff"] = "0"
        srcrect.attrib["yOff"] = "0"
        srcrect.attrib["xSize"] = str(src_dataset.width)
        srcrect.attrib["ySize"] = str(src_dataset.height)
        dstrect = ET.SubElement(source, "DstRect")
        dstrect.attrib["xOff"] = "0"
        dstrect.attrib["yOff"] = "0"
        dstrect.attrib["xSize"] = str(src_dataset.width)
        dstrect.attrib["ySize"] = str(src_dataset.height)

        if src_dataset.options is not None:
            openoptions = ET.SubElement(source, "OpenOptions")
            for ookey, oovalue in src_dataset.options.items():
                ooi = ET.SubElement(openoptions, "OOI")
                ooi.attrib["key"] = str(ookey)
                ooi.text = str(oovalue)

        if nodata_value is not None:
            nodata = ET.SubElement(source, "NODATA")
            nodata.text = str(nodata_value)

    if all(MaskFlags.per_dataset in flags for flags in src_dataset.mask_flag_enums):
        maskband = ET.SubElement(vrtdataset, "MaskBand")
        vrtrasterband = ET.SubElement(maskband, "VRTRasterBand")
        vrtrasterband.attrib["dataType"] = "Byte"

        source = ET.SubElement(vrtrasterband, "SimpleSource")
        sourcefilename = ET.SubElement(source, "SourceFilename")
        sourcefilename.attrib["relativeToVRT"] = "0"
        sourcefilename.attrib["shared"] = "0"
        sourcefilename.text = _parse_path(src_dataset.name).as_vsi()

        sourceband = ET.SubElement(source, "SourceBand")
        sourceband.text = "mask,1"
        sourceproperties = ET.SubElement(source, "SourceProperties")
        sourceproperties.attrib["RasterXSize"] = str(src_dataset.width)
        sourceproperties.attrib["RasterYSize"] = str(src_dataset.height)
        sourceproperties.attrib["dataType"] = "Byte"
        sourceproperties.attrib["BlockYSize"] = str(block_shape[0])
        sourceproperties.attrib["BlockXSize"] = str(block_shape[1])
        srcrect = ET.SubElement(source, "SrcRect")
        srcrect.attrib["xOff"] = "0"
        srcrect.attrib["yOff"] = "0"
        srcrect.attrib["xSize"] = str(src_dataset.width)
        srcrect.attrib["ySize"] = str(src_dataset.height)
        dstrect = ET.SubElement(source, "DstRect")
        dstrect.attrib["xOff"] = "0"
        dstrect.attrib["yOff"] = "0"
        dstrect.attrib["xSize"] = str(src_dataset.width)
        dstrect.attrib["ySize"] = str(src_dataset.height)

    return ET.tostring(vrtdataset).decode("ascii")



"""Common response models."""

from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Field, model_validator
from typing_extensions import Annotated

from titiler.image.settings import iiif_settings


class IIIFSize(BaseModel):
    """
    IIIF Size info

    e.g. `{ "width": 150, "height": 100 }`

    Ref: https://iiif.io/api/image/3.0/#53-sizes
    """

    type: Literal["Size"] = "Size"
    width: int
    height: int


class IIIFTile(BaseModel):
    """
    IIIF Tile info

    e.g `{ "width": 512, "height": 512, "scaleFactors": [ 1, 2, 4, 8, 16 ] }`

    Ref: https://iiif.io/api/image/3.0/#54-tiles
    """

    type: Literal["Tile"] = "Tile"
    scaleFactors: List[int]
    width: int
    height: Optional[int] = None

    @model_validator(mode="before")
    def check_height(cls, values):
        """Check height configuration."""
        if values.get("height", None):
            values["height"] = values["width"]

        return values


class iiifInfo(BaseModel):
    """
    IIIF Info model.

    Based on https://iiif.io/api/image/3.0/#5-image-information

    """

    # The @context tells Linked Data processors how to interpret the image information. If extensions are used then their context definitions should be included in this top-level @context property.
    context: Annotated[
        Literal["http://iiif.io/api/image/3/context.json"],
        Field(alias="@context"),
    ] = "http://iiif.io/api/image/3/context.json"

    # The base URI of the image as defined in URI Syntax, including scheme, server, prefix and identifier without a trailing slash.
    id: str

    # The type for the Image API. The value must be the string ImageService3.
    type: Literal["ImageService3"] = "ImageService3"

    # The URI http://iiif.io/api/image which can be used to determine that the document describes an image service which is a version of the IIIF Image API.
    protocol: Literal["http://iiif.io/api/image"] = "http://iiif.io/api/image"
    # A string indicating the highest compliance level which is fully supported by the service. The value must be one of level0, level1, or level2.
    profile: Literal["level0", "level1", "level2"] = "level2"
    # The width in pixels of the full image, given as an integer.
    width: int
    # The height in pixels of the full image, given as an integer.
    height: int
    # The maximum width in pixels supported for this image.
    maxWidth: Optional[int] = iiif_settings.max_width
    # The maximum height in pixels supported for this image.
    maxHeight: Optional[int] = iiif_settings.max_height
    # The maximum area in pixels supported for this image.
    maxArea: Optional[int] = iiif_settings.max_area

    preferredFormats: Optional[List[str]] = ["png", "jpeg", "webp", "tif", "jp2"]

    # sizes property, which is used to describe preferred height and width combinations for representations of the full image.
    sizes: Optional[List[IIIFSize]] = None

    # tiles property which describes a set of image regions that have a consistent height and width, over a series of resolutions, that can be stitched together visually.
    tiles: Optional[List[IIIFTile]] = None

    # The rights property has the same semantics and requirements as it does in the Presentation API.
    rights: Optional[str] = None

    # https://iiif.io/api/image/3.0/#57-extra-functionality
    # An array of strings that can be used as the quality parameter, in addition to default.
    extraQualities: Optional[List[str]] = None
    # An array of strings that can be used as the format parameter, in addition to the ones specified in the referenced profile.
    extraFormats: Optional[List[str]] = None
    # An array of strings identifying features supported by the service, in addition to the ones specified in the referenced profile.
    extraFeatures: Optional[List[str]] = None

    # https://iiif.io/api/image/3.0/#58-linking-properties
    # TODO: Define models
    partOf: Optional[Dict] = None
    seeAlso: Optional[Dict] = None
    service: Optional[Dict] = None



"""Titiler.image utility functions."""

import math
from typing import List, Optional, Tuple

import numpy
from affine import Affine
from fastapi import HTTPException
from rasterio.warp import reproject
from rio_tiler.models import ImageData


def _percent(x: float, y: float) -> float:
    return (x / 100) * y


def _get_sizes(
    w: int,
    h: int,
    max_width: Optional[int] = None,
    max_height: Optional[int] = None,
    max_area: Optional[int] = None,
) -> Tuple[int, int]:
    """Return Output width/height constrained by environment."""
    # use size constraints if present, else full
    if max_area and max_area < (w * h):
        area_ratio = max_area / (w * h)
        w = int(w * area_ratio)
        h = int(h * area_ratio)

    elif max_width:
        max_height = max_height or max_width
        width, height = w, h

        if w > max_width:
            w = max_width
            h = int(height * max_width / width)

        if h > max_height:
            h = max_height
            w = int(width * max_height / height)

    return w, h


def rotate(img: ImageData, angle: float, expand: bool = False, mirrored: bool = False):
    """Rotate Image.

    Args:
        img (rio_tiler.models.ImageData): ImageData to rotate.
        angle (float): The degrees of clockwise rotation from 0 up to 360.
        expand (bool): Optional expansion flag.  If true, expands the output
           image to make it large enough to hold the entire rotated image.
           If false or omitted, make the output image the same size as the
           input image.  Note that the expand flag assumes rotation around
           the center and no translation.
        mirrored (bool): Optional flag to apply mirroring and then rotation.

    Returns:
        ImageData

    """

    array = img.array
    if mirrored:
        array = numpy.flip(array, axis=2)

    if angle != 0:
        nband = img.count
        nw = img.width
        nh = img.height

        # rotation around image center
        rotated_affine = Affine.rotation(-angle, (nw // 2, nh // 2))

        # Adapted from https://github.com/python-pillow/Pillow/blob/acdb882aae391f29e551a09dc678b153c0c04e5b/src/PIL/Image.py#L2297-L2311
        if expand:
            xx = []
            yy = []
            for x, y in (
                (0, 0),
                (img.width, 0),
                (img.width, img.height),
                (0, img.height),
            ):
                x, y = rotated_affine * (x, y)
                xx.append(x)
                yy.append(y)

            nw = math.ceil(max(xx)) - math.floor(min(xx))
            nh = math.ceil(max(yy)) - math.floor(min(yy))

            rotated_affine = rotated_affine * Affine.translation(
                -(nw - img.width) / 2.0, -(nh - img.height) / 2.0
            )

        # Rotate the data
        data = numpy.zeros((nband, nh, nw), dtype=array.data.dtype)
        _ = reproject(
            array.data,
            data,
            src_crs="epsg:4326",  # Fake CRS
            src_transform=Affine.identity(),
            dst_crs="epsg:4326",  # Fake CRS
            dst_transform=rotated_affine,
        )

        # Rotate the mask
        mask = numpy.zeros((nband, nh, nw), dtype="uint8") + 1
        _ = reproject(
            array.mask * 1,
            mask,
            src_crs="epsg:4326",  # Fake CRS
            src_transform=Affine.identity(),
            dst_crs="epsg:4326",  # Fake CRS
            dst_transform=rotated_affine,
            dst_nodata=1,  # 1=True -> means masked
        )

        array = numpy.ma.MaskedArray(data, mask=mask.astype("bool"))

    return ImageData(
        array,
        assets=img.assets,
        metadata=img.metadata,
        band_names=img.band_names,
        dataset_statistics=img.dataset_statistics,
    )


def image_to_grayscale(img: ImageData) -> ImageData:
    """Convert Image to Grayscale using ITU-R 601-2 luma transform."""
    if img.count == 1:
        return img

    if img.count == 3:
        data = (
            img.data[0] * 299 / 1000
            + img.data[1] * 587 / 1000
            + img.data[2] * 114 / 1000
        )

        data = numpy.ma.MaskedArray(data.astype("uint8"))
        data.mask = ~img.mask.astype("bool")

        return ImageData(
            data,
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
            band_names=["b1"],
            metadata=img.metadata,
        )

    raise HTTPException(
        status_code=400,
        detail=f"Number of band {img.count} for grayscale transformation.",
    )


def image_to_bitonal(img: ImageData) -> ImageData:
    """Convert Image to Bitonal

    All values larger than 127 are set to 255 (white), all other values to 0 (black).
    """
    img = image_to_grayscale(img)
    arr = numpy.where(img.array > 127, 255, 0).astype("uint8")

    return ImageData(
        arr,
        assets=img.assets,
        crs=img.crs,
        bounds=img.bounds,
        band_names=["b1"],
        metadata=img.metadata,
    )


def accept_media_type(accept: str, mediatypes: List[str]) -> Optional[str]:
    """Return MediaType based on accept header and available mediatype.

    Links:
    - https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html
    - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept

    """
    accept_values = {}
    for m in accept.replace(" ", "").split(","):
        values = m.split(";")
        if len(values) == 1:
            name = values[0]
            quality = 1.0
        else:
            name = values[0]
            groups = dict([param.split("=") for param in values[1:]])  # type: ignore
            try:
                q = groups.get("q")
                quality = float(q) if q else 1.0
            except ValueError:
                quality = 0

        # if quality is 0 we ignore encoding
        if quality:
            accept_values[name] = quality

    # Create Preference matrix
    media_preference = {
        v: [n for (n, q) in accept_values.items() if q == v]
        for v in sorted(set(accept_values.values()), reverse=True)
    }

    # Loop through available compression and encoding preference
    for _, pref in media_preference.items():
        for media in mediatypes:
            if media in pref:
                return media

    # If no specified encoding is supported but "*" is accepted,
    # take one of the available compressions.
    if "*" in accept_values and mediatypes:
        return mediatypes[0]

    return None



"""TiTiler-Image FastAPI application."""
import warnings

from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from rasterio.errors import NotGeoreferencedWarning, RasterioIOError
from starlette import status
from starlette.middleware.cors import CORSMiddleware
from starlette_cramjam.middleware import CompressionMiddleware

from titiler.core.errors import DEFAULT_STATUS_CODES, add_exception_handlers
from titiler.core.middleware import CacheControlMiddleware
from titiler.image import __version__ as titiler_image_version
from titiler.image.factory import (
    GeoTilerFactory,
    IIIFFactory,
    LocalTilerFactory,
    MetadataFactory,
)
from titiler.image.settings import api_settings

app = FastAPI(
    title=api_settings.name,
    openapi_url="/api",
    docs_url="/api.html",
    description="""titiler application to work with non-geo images.

---

**Source Code**: <a href="https://github.com/developmentseed/titiler-image" target="_blank">https://github.com/developmentseed/titiler-image</a>

---
    """,
    version=titiler_image_version,
    root_path=api_settings.root_path,
)

warnings.filterwarnings("ignore", category=NotGeoreferencedWarning)

DEFAULT_STATUS_CODES.update(
    {
        RasterioIOError: status.HTTP_404_NOT_FOUND,
        RequestValidationError: status.HTTP_400_BAD_REQUEST,
    }
)

add_exception_handlers(app, DEFAULT_STATUS_CODES)

# Set all CORS enabled origins
if api_settings.cors_origins:
    app.add_middleware(
        CORSMiddleware,
        allow_origins=api_settings.cors_origins,
        allow_credentials=True,
        allow_methods=["GET"],
        allow_headers=["*"],
    )

app.add_middleware(
    CompressionMiddleware,
    minimum_size=0,
    exclude_mediatype={
        "image/jpeg",
        "image/jpg",
        "image/png",
        "image/jp2",
        "image/webp",
    },
)

app.add_middleware(
    CacheControlMiddleware,
    cachecontrol=api_settings.cachecontrol,
    exclude_path={r"/healthz"},
)

meta = MetadataFactory()
app.include_router(meta.router, tags=["Metadata"])

iiif = IIIFFactory(router_prefix="/iiif")
app.include_router(iiif.router, tags=["IIIF"], prefix="/iiif")

image_tiles = LocalTilerFactory(router_prefix="/image")
app.include_router(image_tiles.router, tags=["Local Tiles"], prefix="/image")

geo_tiles = GeoTilerFactory(router_prefix="/geo")
app.include_router(geo_tiles.router, tags=["Geo Tiles"], prefix="/geo")

###############################################################################
# Health Check Endpoint
@app.get(
    "/healthz",
    description="Health Check.",
    summary="Health Check.",
    operation_id="healthCheck",
    tags=["Health Check"],
)
def ping():
    """Health check."""
    return {"ping": "pong!"}



"""titiler.image factories."""

import abc
import urllib.parse
from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Tuple, Type

import jinja2
from fastapi import APIRouter, Depends, HTTPException, Path, Query, params
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from pydantic import conint
from rasterio import windows
from rio_tiler.io import BaseReader, ImageReader
from rio_tiler.models import Info
from starlette.requests import Request
from starlette.responses import (
    HTMLResponse,
    RedirectResponse,
    Response,
    StreamingResponse,
)
from starlette.routing import Match, compile_path, replace_params
from starlette.templating import Jinja2Templates
from typing_extensions import Annotated

from titiler.core.dependencies import (
    BidxExprParams,
    ColorMapParams,
    DefaultDependency,
    HistogramParams,
    ImageParams,
    RescalingParams,
    StatisticsParams,
)
from titiler.core.factory import TilerFactory, img_endpoint_params
from titiler.core.models.mapbox import TileJSON
from titiler.core.models.responses import Statistics
from titiler.core.resources.enums import ImageType, MediaType
from titiler.core.resources.responses import JSONResponse
from titiler.core.routing import EndpointScope
from titiler.image.dependencies import DatasetParams, GCPSParams
from titiler.image.models import iiifInfo
from titiler.image.reader import Reader
from titiler.image.resources.enums import IIIFImageFormat
from titiler.image.settings import iiif_settings
from titiler.image.utils import (
    _get_sizes,
    _percent,
    accept_media_type,
    image_to_bitonal,
    image_to_grayscale,
    rotate,
)

DEFAULT_TEMPLATES = Jinja2Templates(
    directory="",
    loader=jinja2.ChoiceLoader([jinja2.PackageLoader(__package__, "templates")]),
)  # type:ignore


@dataclass  # type: ignore
class BaseFactory(metaclass=abc.ABCMeta):
    """BaseFactory.

    Abstract Base Class for endpoints factories.

    Note: This is a custom version of titiler.core.factory.BaseTilerFactory (striped of most options)

    """

    # FastAPI router
    router: APIRouter = field(default_factory=APIRouter)

    # Router Prefix is needed to find the path for /tile if the TilerFactory.router is mounted
    # with other router (multiple `.../tile` routes).
    # e.g if you mount the route with `/cog` prefix, set router_prefix to cog and
    router_prefix: str = ""

    # add dependencies to specific routes
    route_dependencies: List[Tuple[List[EndpointScope], List[params.Depends]]] = field(
        default_factory=list
    )

    templates: Jinja2Templates = DEFAULT_TEMPLATES

    def __post_init__(self):
        """Post Init: register route and configure specific options."""
        self.register_routes()

        for scopes, dependencies in self.route_dependencies:
            self.add_route_dependencies(scopes=scopes, dependencies=dependencies)

    @abc.abstractmethod
    def register_routes(self):
        """Register Routes."""
        ...

    def url_for(self, request: Request, name: str, **path_params: Any) -> str:
        """Return full url (with prefix) for a specific endpoint."""
        url_path = self.router.url_path_for(name, **path_params)
        base_url = str(request.base_url)
        if self.router_prefix:
            prefix = self.router_prefix.lstrip("/")
            # If we have prefix with custom path param we check and replace them with
            # the path params provided
            if "{" in prefix:
                _, path_format, param_convertors = compile_path(prefix)
                prefix, _ = replace_params(
                    path_format, param_convertors, request.path_params
                )
            base_url += prefix

        return str(url_path.make_absolute_url(base_url=base_url))

    def add_route_dependencies(
        self,
        *,
        scopes: List[EndpointScope],
        dependencies=List[params.Depends],
    ):
        """Add dependencies to routes.

        Allows a developer to add dependencies to a route after the route has been defined.

        """
        for route in self.router.routes:
            for scope in scopes:
                match, _ = route.matches({"type": "http", **scope})
                if match != Match.FULL:
                    continue

                # Mimicking how APIRoute handles dependencies:
                # https://github.com/tiangolo/fastapi/blob/1760da0efa55585c19835d81afa8ca386036c325/fastapi/routing.py#L408-L412
                for depends in dependencies[::-1]:
                    route.dependant.dependencies.insert(  # type: ignore
                        0,
                        get_parameterless_sub_dependant(
                            depends=depends, path=route.path_format  # type: ignore
                        ),
                    )

                # Register dependencies directly on route so that they aren't ignored if
                # the routes are later associated with an app (e.g. app.include_router(router))
                # https://github.com/tiangolo/fastapi/blob/58ab733f19846b4875c5b79bfb1f4d1cb7f4823f/fastapi/applications.py#L337-L360
                # https://github.com/tiangolo/fastapi/blob/58ab733f19846b4875c5b79bfb1f4d1cb7f4823f/fastapi/routing.py#L677-L678
                route.dependencies.extend(dependencies)  # type: ignore


###############################################################################
# Metadata Endpoints Factory
###############################################################################
@dataclass
class MetadataFactory(BaseFactory):
    """Metadata Factory."""

    def register_routes(self):
        """Register Routes."""

        @self.router.get(
            "/info",
            response_model=Info,
            response_model_exclude_none=True,
            response_class=JSONResponse,
            responses={200: {"description": "Return dataset's basic info."}},
        )
        def info(
            src_path: Annotated[
                str,
                Query(description="Dataset URL", alias="url"),
            ],
        ):
            """Return Image metadata."""
            with ImageReader(src_path) as dst:
                return dst.info()

        @self.router.get(
            "/statistics",
            response_class=JSONResponse,
            response_model=Statistics,
            responses={
                200: {
                    "content": {"application/json": {}},
                    "description": "Return dataset's statistics.",
                }
            },
        )
        def statistics(
            src_path: Annotated[
                str,
                Query(description="Dataset URL", alias="url"),
            ],
            layer_params: BidxExprParams = Depends(),
            dataset_params: DatasetParams = Depends(),
            image_params: ImageParams = Depends(),
            stats_params: StatisticsParams = Depends(),
            histogram_params: HistogramParams = Depends(),
        ):
            """Get Dataset statistics."""
            with ImageReader(src_path) as dst:
                return dst.statistics(
                    **layer_params,
                    **image_params,
                    **dataset_params,
                    **stats_params,
                    hist_options={**histogram_params},
                )


###############################################################################
# Local Tiles Endpoints Factory
###############################################################################
@dataclass
class LocalTilerFactory(BaseFactory):
    """Local Tiler Factory."""

    add_viewer: bool = True

    def register_routes(self):
        """Register Routes."""
        self.register_tiles()

        if self.add_viewer:
            self.register_viewer()

    def register_tiles(self):
        """Register Tile routes."""

        @self.router.get(
            "/tilejson.json",
            response_model=TileJSON,
            responses={200: {"description": "Return a TileJSON document"}},
            response_model_exclude_none=True,
        )
        def tilejson(
            request: Request,
            src_path: Annotated[
                str,
                Query(description="Dataset URL", alias="url"),
            ],
            tile_format: Annotated[
                Optional[ImageType],
                Query(description="Output image type. Default is auto."),
            ] = None,
            tile_scale: Annotated[
                Optional[int],
                Query(
                    gt=0,
                    lt=4,
                    description="Tile size scale. 1=256x256, 2=512x512...",
                ),
            ] = None,
            minzoom: Annotated[
                Optional[int],
                Query(description="Overwrite default minzoom."),
            ] = None,
            maxzoom: Annotated[
                Optional[int],
                Query(description="Overwrite default maxzoom."),
            ] = None,
            layer_params: BidxExprParams = Depends(),
            dataset_params: DatasetParams = Depends(),
            rescale: RescalingParams = Depends(),
            color_formula: Annotated[
                Optional[str],
                Query(
                    title="Color Formula",
                    description="rio-color formula (info: https://github.com/mapbox/rio-color)",
                ),
            ] = None,
            colormap: ColorMapParams = Depends(),
            add_mask: Annotated[
                Optional[bool],
                Query(
                    alias="return-mask",
                    description="Add mask to the output data.",
                ),
            ] = None,
        ):
            """return Tilejson doc."""
            route_params: Dict[str, Any] = {
                "z": "{z}",
                "x": "{x}",
                "y": "{y}",
            }

            if tile_scale:
                route_params["scale"] = tile_scale

            if tile_format:
                route_params["format"] = tile_format.value

            tiles_url = self.url_for(request, "tile", **route_params)

            qs_key_to_remove = [
                "tile_format",
                "tile_scale",
                "minzoom",
                "maxzoom",
            ]
            qs = [
                (key, value)
                for (key, value) in request.query_params._list
                if key.lower() not in qs_key_to_remove
            ]
            if qs:
                tiles_url += f"?{urllib.parse.urlencode(qs)}"

            with ImageReader(src_path) as dst:
                return {
                    "bounds": dst.geographic_bounds,
                    "minzoom": minzoom if minzoom is not None else dst.minzoom,
                    "maxzoom": maxzoom if maxzoom is not None else dst.maxzoom,
                    "tiles": [tiles_url],
                }

        @self.router.get("/tiles/{z}/{x}/{y}", **img_endpoint_params)
        @self.router.get("/tiles/{z}/{x}/{y}.{format}", **img_endpoint_params)
        @self.router.get("/tiles/{z}/{x}/{y}@{scale}x", **img_endpoint_params)
        @self.router.get("/tiles/{z}/{x}/{y}@{scale}x.{format}", **img_endpoint_params)
        def tile(
            z: Annotated[
                int,
                Path(description="Identifier (Z) selecting one of the scales."),
            ],
            x: Annotated[
                int,
                Path(description="Column (X) index of the tile."),
            ],
            y: Annotated[
                int,
                Path(description="Row (Y) index of the tile."),
            ],
            src_path: Annotated[str, Query(description="Dataset URL", alias="url")],
            scale: Annotated[
                Optional[conint(gt=0, le=4)], "Tile size scale. 1=256x256, 2=512x512..."
            ] = None,
            format: Annotated[
                ImageType,
                "Default will be automatically defined if the output image needs a mask (png) or not (jpeg).",
            ] = None,
            layer_params: BidxExprParams = Depends(),
            dataset_params: DatasetParams = Depends(),
            rescale: RescalingParams = Depends(),
            color_formula: Annotated[
                Optional[str],
                Query(
                    title="Color Formula",
                    description="rio-color formula (info: https://github.com/mapbox/rio-color)",
                ),
            ] = None,
            colormap: ColorMapParams = Depends(),
            add_mask: Annotated[
                Optional[bool],
                Query(
                    alias="return-mask",
                    description="Add mask to the output data.",
                ),
            ] = None,
        ):
            """Tile in Local TMS."""
            tilesize = scale * 256 if scale is not None else 256

            with ImageReader(src_path) as dst:
                image = dst.tile(
                    x,
                    y,
                    z,
                    tilesize=tilesize,
                    **layer_params,
                    **dataset_params,
                )
                dst_colormap = getattr(dst, "colormap", None)

            if rescale:
                image.rescale(rescale)

            if color_formula:
                image.apply_color_formula(color_formula)

            if cmap := colormap or dst_colormap:
                image = image.apply_colormap(cmap)

            if not format:
                format = ImageType.jpeg if image.mask.all() else ImageType.png

            content = image.render(
                add_mask=add_mask if add_mask is not None else True,
                img_format=format.driver,
                **format.profile,
            )

            return Response(content, media_type=format.mediatype)

    def register_viewer(self):
        """Register Viewer route."""

        @self.router.get("/viewer", response_class=HTMLResponse)
        def image_viewer(
            request: Request,
            src_path: Annotated[str, Query(description="Dataset URL", alias="url")],
            tile_format: Annotated[
                Optional[ImageType],
                Query(description="Output image type. Default is auto."),
            ] = None,
            tile_scale: Annotated[
                Optional[int],
                Query(
                    gt=0,
                    lt=4,
                    description="Tile size scale. 1=256x256, 2=512x512...",
                ),
            ] = None,
            minzoom: Annotated[
                Optional[int],
                Query(description="Overwrite default minzoom."),
            ] = None,
            maxzoom: Annotated[
                Optional[int],
                Query(description="Overwrite default maxzoom."),
            ] = None,
            layer_params: BidxExprParams = Depends(),
            dataset_params: DatasetParams = Depends(),
            rescale: RescalingParams = Depends(),
            color_formula: Annotated[
                Optional[str],
                Query(
                    title="Color Formula",
                    description="rio-color formula (info: https://github.com/mapbox/rio-color)",
                ),
            ] = None,
            colormap: ColorMapParams = Depends(),
            add_mask: Annotated[
                Optional[bool],
                Query(
                    alias="return-mask",
                    description="Add mask to the output data.",
                ),
            ] = None,
        ):
            """Return Simple Image viewer."""
            tilejson_url = self.url_for(request, "tilejson")
            if request.query_params._list:
                tilejson_url += f"?{urllib.parse.urlencode(request.query_params._list)}"

            return self.templates.TemplateResponse(
                name="local.html",
                context={
                    "request": request,
                    "tilejson_endpoint": tilejson_url,
                },
                media_type=MediaType.html.value,
            )


###############################################################################
# IIIF Endpoints Factory
###############################################################################
@dataclass
class IIIFFactory(BaseFactory):
    """IIIF Factory.

    Specification: https://iiif.io/api/image/3.0/
    """

    def register_routes(self):
        """Register Routes."""
        self.register_image_api()

    def register_image_api(self):  # noqa: C901
        """Register IIIF Image API routes."""

        @self.router.get(
            "/{identifier:path}/info.json",
            response_model=iiifInfo,
            response_model_exclude_none=True,
            responses={
                200: {
                    "description": "Image Information Request",
                    "content": {
                        "application/json": {},
                        "application/ld+json": {},
                    },
                },
            },
        )
        def iiif_info(
            request: Request,
            identifier: Annotated[
                str,
                Path(description="The identifier of the requested image."),
            ],
        ):
            """Image Information Request."""
            output_type = accept_media_type(
                request.headers.get("accept", ""),
                ["application/json", "application/ld+json"],
            )
            url_path = self.url_for(
                request,
                "iiif_baseuri",
                identifier=urllib.parse.quote_plus(identifier, safe=""),
            )

            identifier = urllib.parse.unquote(identifier)
            with ImageReader(identifier) as dst:
                # TODO: If overviews:
                # Set Sizes
                # Set Tiles (using min/max zooms)
                info = iiifInfo(
                    id=url_path,
                    width=dst.dataset.width,
                    height=dst.dataset.height,
                )

                if output_type == "application/ld+json":
                    return StreamingResponse(
                        iter((info.model_dump_json(exclude_none=True) + "\n",)),
                        media_type='application/ld+json;profile="http://iiif.io/api/image/3/context.json"',
                    )

            return info

        @self.router.get(
            "/{identifier:path}/{region}/{size}/{rotation}/{quality}.{format}",
            **img_endpoint_params,
        )
        def iiif_image(  # noqa: C901
            identifier: Annotated[
                str,
                Path(description="The identifier of the requested image."),
            ],
            region: Annotated[
                str,
                Path(
                    description="The region parameter defines the rectangular portion of the underlying image content to be returned."
                ),
            ],
            size: Annotated[
                str,
                Path(
                    description="The size parameter specifies the dimensions to which the extracted region, which might be the full image, is to be scaled."
                ),
            ],
            rotation: Annotated[
                str,
                Path(
                    description="The rotation parameter specifies mirroring and rotation",
                ),
            ],
            quality: Annotated[
                Literal["color", "gray", "bitonal", "default"],
                Path(
                    description="The quality parameter determines whether the image is delivered in color, grayscale or black and white.",
                ),
            ],
            format: Annotated[
                IIIFImageFormat,
                Path(
                    description="The format of the returned image is expressed as a suffix, mirroring common filename extensions, at the end of the URI."
                ),
            ],
            # TiTiler Extension
            layer_params: BidxExprParams = Depends(),
            dataset_params: DatasetParams = Depends(),
            rescale: RescalingParams = Depends(),
            color_formula: Annotated[
                Optional[str],
                Query(
                    title="Color Formula",
                    description="rio-color formula (info: https://github.com/mapbox/rio-color)",
                ),
            ] = None,
            colormap: ColorMapParams = Depends(),
            add_mask: Annotated[
                Optional[bool],
                Query(
                    alias="return-mask",
                    description="Add mask to the output data.",
                ),
            ] = None,
        ):
            """IIIF Image Request.

            ref: https://iiif.io/api/image/3.0

            """
            identifier = urllib.parse.unquote(identifier)
            with ImageReader(identifier) as dst:
                dst_width = dst.dataset.width
                dst_height = dst.dataset.height

                #################################################################################
                # REGION
                # full, square, x,y,w,h, pct:x,y,w,h
                #################################################################################
                window = windows.Window(
                    col_off=0, row_off=0, width=dst_width, height=dst_height
                )
                if region == "full":
                    # The full image is returned, without any cropping.
                    pass

                elif region == "square":
                    # The region is defined as an area where the width and height are both equal to the length of the shorter dimension of the full image.
                    # The region may be positioned anywhere in the longer dimension of the full image at the server’s discretion, and centered is often a reasonable default.
                    min_size = min(dst_width, dst_height)
                    x_off = (dst_width - min_size) // 2
                    y_off = (dst_height - min_size) // 2
                    window = windows.Window(
                        col_off=x_off, row_off=y_off, width=min_size, height=min_size
                    )

                elif region.startswith("pct:"):
                    # The region to be returned is specified as a sequence of percentages of the full image’s dimensions,
                    # as reported in the image information document.
                    # Thus, x represents the number of pixels from the 0 position on the horizontal axis, calculated as a percentage of the reported width.
                    # w represents the width of the region, also calculated as a percentage of the reported width.
                    # The same applies to y and h respectively.
                    x, y, w, h = list(map(float, region.replace("pct:", "").split(",")))
                    if max(x, y, w, h) > 100 or min(x, y, w, h) < 0:
                        raise HTTPException(
                            status_code=400,
                            detail=f"Invalid Region parameter: {region}.",
                        )

                    x = round(_percent(dst_width, x))
                    y = round(_percent(dst_height, y))
                    w = round(_percent(dst_width, w))
                    h = round(_percent(dst_height, h))

                    # Service should return an image cropped at the image’s edge, rather than adding empty space.
                    w = dst_width - x if w + x > dst_width else w
                    h = dst_height - y if h + y > dst_height else h

                    window = windows.Window(col_off=x, row_off=y, width=w, height=h)

                elif len(region.split(",")) == 4:
                    # The region of the full image to be returned is specified in terms of absolute pixel values.
                    # The value of x represents the number of pixels from the 0 position on the horizontal axis.
                    # The value of y represents the number of pixels from the 0 position on the vertical axis.
                    # Thus the x,y position 0,0 is the upper left-most pixel of the image. w represents
                    # the width of the region and h represents the height of the region in pixels.
                    x, y, w, h = list(map(float, region.split(",")))

                    # Service should return an image cropped at the image’s edge, rather than adding empty space.
                    w = dst_width - x if w + x > dst_width else w
                    h = dst_height - y if h + y > dst_height else h

                    try:
                        window = windows.Window(col_off=x, row_off=y, width=w, height=h)
                    except ValueError as e:
                        raise HTTPException(
                            status_code=400,
                            detail=f"Invalid Region parameter: {region}.",
                        ) from e

                else:
                    raise HTTPException(
                        status_code=400, detail=f"Invalid Region parameter: {region}."
                    )

                if (
                    window.width <= 0
                    or window.height <= 0
                    or window.col_off > dst_width
                    or window.row_off > dst_height
                ):
                    raise HTTPException(
                        status_code=400, detail=f"Invalid Region parameter: {region}."
                    )

                #################################################################################
                # SIZE
                # Formats are: w, ,h w,h pct:p !w,h full max ^w, ^,h ^w,h
                #################################################################################
                out_width, out_height = window.width, window.height
                aspect_ratio = out_width / out_height

                if size == "max":
                    # max: The extracted region is returned at the maximum size available, but will not be upscaled.
                    # The resulting image will have the pixel dimensions of the extracted region,
                    # unless it is constrained to a smaller size by maxWidth, maxHeight, or maxArea
                    pass

                elif size == "^max":
                    # ^max: The extracted region is scaled to the maximum size permitted by maxWidth, maxHeight, or maxArea.
                    # If the resulting dimensions are greater than the pixel width and height of the extracted region, the extracted region is upscaled.
                    if aspect_ratio > 1:
                        out_width = (
                            max(out_width, iiif_settings.max_width)
                            if iiif_settings.max_width
                            else out_width
                        )
                        out_height = round(out_width / aspect_ratio)
                    else:
                        out_height = (
                            max(out_height, iiif_settings.max_height)
                            if iiif_settings.max_height
                            else out_height
                        )
                        out_width = round(aspect_ratio * out_height)

                elif size.startswith("pct:"):
                    # pct:n: The width and height of the returned image is scaled to n percent of the width and height of the extracted region.
                    # The value of n must not be greater than 100.
                    pct_size = float(size.replace("pct:", ""))
                    if pct_size > 100 or pct_size <= 0:
                        raise HTTPException(
                            status_code=400,
                            detail=f"Invalid Size parameter: {size} (must be between 0 and 100).",
                        )

                    out_width = round(_percent(out_width, pct_size))
                    out_height = round(_percent(out_height, pct_size))

                elif size.startswith("^pct:"):
                    # ^pct:n: The width and height of the returned image is scaled to n percent of the width and height of the extracted region.
                    # For values of n greater than 100, the extracted region is upscaled.
                    pct_size = float(size.replace("^pct:", ""))
                    if pct_size <= 0:
                        raise HTTPException(
                            status_code=400,
                            detail=f"Invalid Size parameter: {size} (must be greater than 0).",
                        )

                    out_width = round(_percent(out_width, pct_size))
                    out_height = round(_percent(out_height, pct_size))

                elif "," in size:
                    sizes = size.split(",")

                    if size.startswith("^!"):
                        # ^!w,h	The extracted region is scaled so that the width and height of the returned image are not greater than w and h, while maintaining the aspect ratio.
                        # The returned image must be as large as possible but not larger than w, h, or server-imposed limits.
                        max_width, max_height = list(
                            map(int, size.replace("^!", "").split(","))
                        )
                        if aspect_ratio > 1:
                            out_width = max_width
                            out_height = round(out_width / aspect_ratio)
                        else:
                            out_height = max_height
                            out_width = round(aspect_ratio * out_height)

                    elif size.startswith("!"):
                        # !w,h	The extracted region is scaled so that the width and height of the returned image are not greater than w and h, while maintaining the aspect ratio.
                        # The returned image must be as large as possible but not larger than the extracted region, w or h, or server-imposed limits.
                        max_width, max_height = list(
                            map(int, size.replace("!", "").split(","))
                        )

                        if aspect_ratio > 1:
                            out_width = max_width
                            out_height = round(out_width / aspect_ratio)
                        else:
                            out_height = max_height
                            out_width = round(aspect_ratio * out_height)

                        if out_width > window.width or out_height > window.height:
                            raise HTTPException(
                                status_code=400,
                                detail=f"Invalid 'h,w' parameter: {size} (greater than region height {window.width},{window.height}).",
                            )

                    elif size.startswith("^"):
                        sizes = size.replace("^", "").split(",")

                        if size.endswith(","):
                            # ^w,: The extracted region should be scaled so that the width of the returned image is exactly equal to w.
                            # If w is greater than the pixel width of the extracted region, the extracted region is upscaled.
                            out_width = int(sizes[0])
                            out_height = round(out_width / aspect_ratio)

                        elif size.startswith("^,"):
                            # ^,h: The extracted region should be scaled so that the height of the returned image is exactly equal to h. If h is greater than the pixel height of the extracted region, the extracted region is upscaled.
                            out_height = int(sizes[1])
                            out_width = round(aspect_ratio * out_height)

                        elif sizes[0] and sizes[1]:
                            # ^w,h:	The width and height of the returned image are exactly w and h.
                            # The aspect ratio of the returned image may be significantly different than the extracted region, resulting in a distorted image.
                            # If w and/or h are greater than the corresponding pixel dimensions of the extracted region, the extracted region is upscaled.
                            out_width, out_height = list(map(int, sizes))

                    elif size.endswith(","):
                        # w,: The extracted region should be scaled so that the width of the returned image is exactly equal to w.
                        # The value of w must not be greater than the width of the extracted region.
                        out_width = int(sizes[0])
                        if out_width > window.width:
                            raise HTTPException(
                                status_code=400,
                                detail=f"Invalid 'w' parameter: {out_width} (greater than region width {window.width}).",
                            )
                        out_height = round(out_width / aspect_ratio)

                    elif size.startswith(","):
                        # ,h: The extracted region should be scaled so that the height of the returned image is exactly equal to h.
                        # The value of h must not be greater than the height of the extracted region.
                        out_height = int(sizes[1])
                        if out_height > window.height:
                            raise HTTPException(
                                status_code=400,
                                detail=f"Invalid 'h' parameter: {out_height} (greater than region height {window.height}).",
                            )
                        out_width = round(aspect_ratio * out_height)

                    elif sizes[0] and sizes[1]:
                        # w,h: The width and height of the returned image are exactly w and h.
                        # The aspect ratio of the returned image may be significantly different than the extracted region, resulting in a distorted image.
                        # The values of w and h must not be greater than the corresponding pixel dimensions of the extracted region.
                        out_width, out_height = list(map(int, sizes))
                        if out_width > window.width or out_height > window.height:
                            raise HTTPException(
                                status_code=400,
                                detail=f"Invalid 'h,w' parameter: {size} (greater than region height {window.width},{window.height}).",
                            )

                    else:
                        raise HTTPException(
                            status_code=400, detail=f"Invalid Size parameter: {size}."
                        )

                else:
                    raise HTTPException(
                        status_code=400, detail=f"Invalid Size parameter: {size}."
                    )

                out_width, out_height = _get_sizes(
                    out_width,
                    out_height,
                    max_width=iiif_settings.max_width,
                    max_height=iiif_settings.max_height,
                    max_area=iiif_settings.max_area,
                )

                if out_width <= 1 or out_height <= 1:
                    raise HTTPException(
                        status_code=400,
                        detail=f"Invalid Size parameter: {size} resulting in size <=1 ({out_width},{out_height}).",
                    )

                image = dst.read(
                    window=window,
                    width=int(out_width),
                    height=int(out_height),
                    **layer_params,
                    **dataset_params,
                )
                dst_colormap = getattr(dst, "colormap", None)

            #################################################################################
            # ROTATION
            # Formats are: n, !n
            #################################################################################
            try:
                rot = float(rotation.replace("!", ""))
                if rot < 0 or rot > 360:
                    raise ValueError("Invalid rotation value")

            except (ValueError) as e:
                raise HTTPException(
                    status_code=400,
                    detail=f"Invalid rotation parameter: {rotation}.",
                ) from e

            image = rotate(image, rot, expand=True, mirrored=rotation.startswith("!"))

            if rescale:
                image.rescale(rescale)

            if color_formula:
                image.apply_color_formula(color_formula)

            #################################################################################
            # QUALITY
            # one of default, color, gray or bitonal
            #################################################################################
            if quality == "gray":
                colormap = dst_colormap = None
                image = image_to_grayscale(image)

            elif quality == "bitonal":
                colormap = dst_colormap = None
                image = image_to_bitonal(image)

            if cmap := colormap or dst_colormap:
                image = image.apply_colormap(cmap)

            content = image.render(
                add_mask=add_mask if add_mask is not None else True,
                img_format=format.driver,
                **format.profile,
            )
            return Response(content, media_type=format.mediatype)

        @self.router.get(
            "/{identifier:path}",
            response_class=RedirectResponse,
            include_in_schema=False,
        )
        def iiif_baseuri(
            request: Request,
            identifier: Annotated[
                str, Path(description="The identifier of the requested image.")
            ],
        ):
            """Return Simple IIIF viewer.

            ref: https://iiif.io/api/image/3.0/#2-uri-syntax
            When the base URI is dereferenced, the interaction should result in the image information document.
            It is recommended that the response be a 303 status redirection to the image information document’s URI.
            Implementations may also exhibit other behavior for the base URI beyond the scope of this specification
            in response to HTTP request headers and methods.

            """
            url = self.url_for(request, "iiif_info", identifier=identifier)
            output_type = accept_media_type(
                request.headers.get("accept", ""),
                ["text/html"],
            )
            if output_type:
                return self.templates.TemplateResponse(
                    name="iiif.html",
                    context={
                        "request": request,
                        "info_endpoint": url,
                    },
                    media_type=MediaType.html.value,
                )

            return RedirectResponse(url)


###############################################################################
# Geo Tiler Factory
###############################################################################
@dataclass
class GeoTilerFactory(TilerFactory):
    """Like Tiler Factory but with less endpoints."""

    reader: Type[BaseReader] = Reader

    reader_dependency: Type[DefaultDependency] = GCPSParams

    # Rasterio Dataset Options (nodata, unscale, resampling)
    dataset_dependency: Type[DefaultDependency] = DatasetParams

    def register_routes(self):
        """This Method register routes to the router."""
        self.tile()
        self.tilejson()

        if self.add_viewer:
            self.map_viewer()



"""Titiler.image Enums."""

from enum import Enum
from types import DynamicClassAttribute

from rio_tiler.profiles import img_profiles

from titiler.core.resources.enums import ImageDriver, MediaType


class IIIFImageFormat(str, Enum):
    """Available Output image type."""

    jpg = "jpg"
    tif = "tif"
    png = "png"
    gif = "gif"
    jp2 = "jp2"
    # pdf = "pdf" Not Available

    @DynamicClassAttribute
    def profile(self):
        """Return rio-tiler image default profile."""
        return img_profiles.get(self._name_, {})

    @DynamicClassAttribute
    def driver(self):
        """Return rio-tiler image default profile."""
        return ImageDriver[self._name_].value

    @DynamicClassAttribute
    def mediatype(self):
        """Return image media type."""
        return MediaType[self._name_].value
